'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _async = require('async');

var _async2 = _interopRequireDefault(_async);

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _fbWatchman = require('fb-watchman');

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _fsAccurency = require('./utils/fsAccurency');

var _fsAccurency2 = _interopRequireDefault(_fsAccurency);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var debug = (0, _debug2.default)('watchman:connector');

var WatchmanConnector = function (_EventEmitter) {
  _inherits(WatchmanConnector, _EventEmitter);

  function WatchmanConnector() {
    var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : { aggregateTimeout: 200, projectPath: '' };

    _classCallCheck(this, WatchmanConnector);

    var _this = _possibleConstructorReturn(this, (WatchmanConnector.__proto__ || Object.getPrototypeOf(WatchmanConnector)).call(this));

    _this.aggregatedChanges = [];
    _this.connected = false;
    _this.fileTimes = {};
    _this.paused = true;
    _this.timeoutRef = 0;
    _this.initialScan = true;
    _this.initialScanRemoved = false;
    _this.initialScanQueue = new Set();

    _this._onSubscription = function (resp) {
      debug('received subscription: %O', resp);
      if (resp.subscription === 'webpack_subscription') {
        _this.lastClock = resp.clock;
        resp.files.forEach(function (file) {
          var filePath = _path2.default.join(_this.options.projectPath, file.name);
          var mtime = !file.exists ? null : +file.mtime_ms;

          _this._setFileTime(filePath, mtime);

          if (_this.initialScan) {
            if (mtime) {
              _this.initialScanQueue.add({ name: filePath, mtime: mtime });
            } else {
              _this.initialScanRemoved = true;
            }
            return;
          }

          if (_this.paused || !file.exists) return;

          _this._handleEvents(filePath, mtime);
        });
      }
    };

    _this._onTimeout = function () {
      _this.timeoutRef = 0;
      var changes = _this.aggregatedChanges;
      _this.aggregatedChanges = [];

      _this.emit('aggregated', changes, _this.lastClock);
    };

    if (!options.projectPath) throw new Error('projectPath is missing for WatchmanPlugin');

    _this.options = options;
    return _this;
  }

  /**
   * `since` has to be either a string with a watchman clock value, or a number
   * which is then treated as a timestamp in milliseconds
   */


  _createClass(WatchmanConnector, [{
    key: 'watch',
    value: function watch(files, dirs, since, done) {
      var _this2 = this;

      debug('watch() called, current connection status: ' + (this.connected ? 'connected' : 'disconnected'));
      this.paused = false;

      if (this.connected) return;

      var allFiles = files.concat(dirs);

      Promise.all([new Promise(function (resolve, reject) {
        _this2._startWatch(allFiles, since, function (err) {
          return err ? reject(err) : resolve();
        });
      }), new Promise(function (resolve) {
        _this2._doInitialScan(allFiles, resolve);
      })]).catch(function (err) {
        throw err;
      }).then(function () {
        return done ? done() : null;
      });
    }
  }, {
    key: 'getTimes',
    value: function getTimes() {
      return this.fileTimes;
    }
  }, {
    key: 'close',
    value: function close() {
      debug('close() called');
      this.paused = true;
      if (this.timeoutRef) clearTimeout(this.timeoutRef);
      this.removeAllListeners();

      var client = this.client;
      if (client) {
        client.removeListener('subscription', this._onSubscription);
        client.command(['unsubscribe', this.options.projectPath, 'webpack_subscription']);
        client.end();
        this.client = null;
      }
    }
  }, {
    key: 'pause',
    value: function pause() {
      debug('pause() called');
      this.paused = true;
      if (this.timeoutRef) clearTimeout(this.timeoutRef);
    }
  }, {
    key: '_startWatch',
    value: function _startWatch(files, since, done) {
      var _this3 = this;

      var client = this._getClientInstance();

      client.capabilityCheck({ optional: [], required: ['cmd-watch-project', 'relative_root'] }, function (capabilityErr) {
        /* istanbul ignore if: cannot happen in tests */
        if (capabilityErr) {
          done(capabilityErr);
          return;
        }
        debug('watchman capabilityCheck() successful');

        // Initiate the watch
        client.command(['watch-project', _this3.options.projectPath], function (watchError, watchResponse) {
          /* istanbul ignore if: cannot happen in tests */
          if (watchError) {
            done(watchError);
            return;
          }
          debug('watchman command watch-project successful');

          /* istanbul ignore if: cannot happen in tests */
          if (watchResponse.warning) {
            console.warn('warning: ', watchResponse.warning); // eslint-disable-line no-console
          }

          var sub = {
            expression: ['allof', ['name', files.map(function (file) {
              return _path2.default.relative(_this3.options.projectPath, file);
            }), 'wholename']],
            fields: ['name', 'mtime_ms', 'exists'],
            since: typeof since === 'string' ? since : Math.floor(since / 1000),
            relative_root: watchResponse.relative_path
          };

          client.on('subscription', _this3._onSubscription);

          debug('watchman command subscription data: ', sub);

          client.command(['subscribe', watchResponse.watch, 'webpack_subscription', sub], function (subscribeError) {
            /* istanbul ignore if: cannot happen in testsn */
            if (subscribeError) {
              done(subscribeError);
              return;
            }
            debug('watchman command subscribe successful');
            done();
          });
        });
      });
    }
  }, {
    key: '_setFileTime',
    value: function _setFileTime(file, mtime) {
      this.fileTimes[file] = mtime + _fsAccurency2.default.get();
    }
  }, {
    key: '_handleEvents',
    value: function _handleEvents(filePath, mtime) {
      this.emit('change', filePath, mtime);

      this._handleAggregated(filePath);
    }
  }, {
    key: '_handleAggregated',
    value: function _handleAggregated(file) {
      if (this.timeoutRef) clearTimeout(this.timeoutRef);

      if (this.aggregatedChanges.indexOf(file) < 0) {
        this.aggregatedChanges.push(file);
      }

      this.timeoutRef = setTimeout(this._onTimeout, this.options.aggregateTimeout);
    }
  }, {
    key: '_getClientInstance',
    value: function _getClientInstance() {
      var _this4 = this;

      if (!this.client) {
        var client = new _fbWatchman.Client();
        client.on('connect', function () {
          _this4.connected = true;
        });
        client.on('end', function () {
          _this4.connected = false;
        });

        this.client = client;
      }

      return this.client;
    }
  }, {
    key: '_doInitialScan',
    value: function _doInitialScan(files, done) {
      var _this5 = this;

      debug('starting initial file scan');
      _async2.default.eachLimit(files, 500, function (file, callback) {
        _fs2.default.stat(file, function (err, stat) {
          if (err) {
            callback(err);
            return;
          }

          var mtime = +stat.mtime;
          _fsAccurency2.default.revalidate(mtime);

          _this5._setFileTime(file, mtime);
          callback();
        });
      }, function () {
        _this5.initialScan = false;
        debug('initial file scan finished');

        if (_this5.initialScanQueue.size > 0) {
          var file = Array.from(_this5.initialScanQueue)[_this5.initialScanQueue.size - 1];
          _this5._handleEvents(file.name, file.mtime);
        }

        if (_this5.initialScanRemoved) {
          _this5.initialScanRemoved = false;
          _this5.emit('remove');
        }

        _this5.initialScanQueue.clear();
        done();
      });
    }
  }]);

  return WatchmanConnector;
}(_events2.default);

exports.default = WatchmanConnector;